#![allow(dead_code)]

use sparse::Sparse;
use pool::Pool;
use trie::Trie;
use crate::{
    system::filesystem::FileSystem,
    util::Chain,
};

pub mod connection; use connection::Connection;
pub mod user; use user::User;
pub mod authentication; use authentication::Authentication;
pub mod session; use session::{Session, SessionToken};
pub mod context;

pub struct App {
    pub filesystem:FileSystem,

    pub connections:Pool<Connection>,

    pub users:Pool<User>,
    pub user_next:u32,
    pub user_id:Sparse<usize>,
    pub user_handle:Trie<u32>,
    pub salts:Sparse<[u8; 16]>,

    pub auths:Trie<Authentication>,
    pub sessions:Trie<Session>,

    pub session_time:Chain<SessionToken>,
}
impl App {
    pub fn init() -> Result<Self, ()>
    {
        if let Ok(mut filesystem) = FileSystem::init() {

            // Load salts
            let mut salts = Sparse::new();
            let salt_count = filesystem.salt_count()?;
            for id in 0..salt_count {
                let salt = filesystem.salt_fetch(id as u32).unwrap();
                salts.set(id as isize, salt);
            }

            // Load handles
            let mut user_handle = Trie::new();
            let handle_count = filesystem.handle_count()?;
            for id in 0..handle_count {
                let (handle, user_id) = filesystem.handle_fetch(id as u32).unwrap();
                println!("got: {} = {}", handle, user_id);
                user_handle.set(handle.as_bytes(), user_id);
            }

            Ok(Self {
                filesystem:filesystem,

                connections:Pool::new(),

                users:Pool::new(),
                user_next:0,
                user_id:Sparse::new(),
                user_handle:Trie::new(),
                salts,

                auths:Trie::new(),
                sessions:Trie::new(),

                session_time:Chain::new(),
            })
        } else {
            Err(())
        }
    }
}
