use std::{
    fs::{self, File}, io::{Read, Seek, SeekFrom, Write}, path::Path
};

use crate::{
    app::session::Session,
    app::user::User,
    util::pack::{pack_u32, unpack_u32}
};

const HANDLE_BUCKET_MASK :u32 = 0xFF;
const HANDLE_BUCKET_SIZE :u32 = HANDLE_BUCKET_MASK + 1;

const DIR_DATA :&str = "data";
const DIR_HANDLE :&str = const_format::formatcp!("{}/h", DIR_DATA);
const DIR_SESSION :&str = const_format::formatcp!("{}/s", DIR_DATA);
const DIR_USER :&str = const_format::formatcp!("{}/u", DIR_DATA);

const INDEX_HANDLE :&str = const_format::formatcp!("{}/i.bin", DIR_HANDLE);
const INDEX_SESSION :&str = const_format::formatcp!("{}/i.bin", DIR_SESSION);
const INDEX_USER :&str = const_format::formatcp!("{}/i.bin", DIR_USER);

pub const FILE_SALT :&str = const_format::formatcp!("{}/x.bin", DIR_DATA);

pub struct FileSystem {
    index_handle:File,
    index_session:File,
    index_user:File,

    table_salt:File,
}
impl FileSystem {
    pub fn init() -> Result<Self, std::io::Error>
    {
        //
        // TEMPORARY: REMOVE AFTER TESTING
        //
        //fs::remove_dir_all(DIR_DATA).ok();
        
        // Initialize filesystem if does not exist.
        //
        // Note: does not currently check for corruption.
        //
        if !Path::new(DIR_DATA).exists() {
            fs::create_dir(DIR_DATA)?;
            fs::create_dir(DIR_HANDLE)?;
            fs::create_dir(DIR_SESSION)?;
            fs::create_dir(DIR_USER)?;

            fs::write(INDEX_HANDLE, pack_u32(0))?;
            fs::write(INDEX_SESSION, pack_u32(0))?;
            fs::write(INDEX_USER, pack_u32(0))?;

            fs::write(FILE_SALT, pack_u32(0))?;
        }

        let index_handle = File::options()
            .read(true)
            .write(true)
            .open(INDEX_HANDLE)?;
        let index_session = File::options()
            .read(true)
            .write(true)
            .open(INDEX_SESSION)?;
        let index_user = File::options()
            .read(true)
            .write(true)
            .open(INDEX_USER)?;

        let table_salt = File::options()
            .read(true)
            .write(true)
            .open(FILE_SALT)?;

        Ok(Self {
            index_handle,
            index_session,
            index_user,

            table_salt,
        })
    }


    pub fn session_store(&mut self, _session:&Session) -> Result<(),()>
    {
        Err(())
    }


    pub fn user_store(&mut self, _user:&User) -> Result<(),()>
    {
        Err(())
    }


    pub fn handle_store(&mut self, handle:&String, user_id:u32) -> Result<u32,()>
    // Add a salt to store.
    //
    {
        let size = self.handle_count()? as u32;
        let data = handle.as_bytes();
        let length = data.len();

        let bucket_index = size & !HANDLE_BUCKET_MASK;
        let record_index = size & HANDLE_BUCKET_MASK;
        let record_offset = 12 * record_index as u64;

        // Update size record
        self.index_handle.seek(SeekFrom::Start(0)).map_err(|_| ())?;
        self.index_handle.write(&pack_u32(size + 1)).map_err(|_| ())?;

        // Create bucket file if not exists
        let bucket_path = Path::new(DIR_HANDLE).join(format!("{:x}.bin", bucket_index));
        if !bucket_path.exists() {
            fs::write(bucket_path.clone(), vec![0u8; 12 * HANDLE_BUCKET_SIZE as usize]).map_err(|_| ())?;
        }

        // Open bucket file for record
        if let Ok(mut file) = File::options().read(true).write(true).open(bucket_path) {
            let offset = file.seek(SeekFrom::End(0)).map_err(|_| ())?;

            // Write record header to table
            let buffer_header = [
                pack_u32(offset as u32),
                pack_u32(length as u32),
                pack_u32(user_id),
            ].concat();

            file.seek(SeekFrom::Start(record_offset)).map_err(|_| ())?;
            file.write(&buffer_header).map_err(|_| ())?;


            // Write handle data to end of file
            file.seek(SeekFrom::End(0)).map_err(|_| ())?;
            file.write(data).map_err(|_| ())?;

            Ok(size)
        } else { Err(()) }
    }

    pub fn handle_fetch(&mut self, id:u32) -> Result<(String, u32),()>
    // Retrieve a salt from store.
    //
    {
        let size = self.salt_count()? as u32;
        let mut buffer = Vec::new();
       
        // Get location of handle by index
        if id < size {
            let bucket_index = id & !HANDLE_BUCKET_MASK;
            let record_index = id & HANDLE_BUCKET_MASK;
            let record_offset = 12 * record_index as u64;

            let path = Path::new(DIR_HANDLE)
                .join(format!("{:x}.bin", bucket_index));

            // Read bucket file for record
            if let Ok(mut file) = File::open(path) {

                // Get record offset, length, and user association
                let mut buffer_header = [0u8; 12];
                file.seek(SeekFrom::Start(record_offset)).map_err(|_| ())?;
                if file.read_exact(&mut buffer_header).is_ok() {
                    let offset = unpack_u32(&buffer_header, &mut 0);
                    let length = unpack_u32(&buffer_header, &mut 4);
                    let user_id = unpack_u32(&buffer_header, &mut 8);

                    if offset != 0 {
                        buffer.resize(length as usize, 0);

                        // Read handle data from offset
                        file.seek(SeekFrom::Start(offset as u64)).map_err(|_| ())?;
                        match file.read_exact(&mut buffer) {
                            Ok(_) => {
                                Ok((String::from_utf8(buffer).map_err(|_| ())?, user_id))
                            }
                            Err(e) => { println!("e: {}", e.to_string()); Err(()) }
                        }
                    } else { Err(()) }
                } else { Err(()) }
            } else { Err(()) }
        } else { Err(()) }
    }

    pub fn handle_count(&mut self) -> Result<usize, ()>
    // Get number of salts in store.
    //
    {
        Self::get_header_size(&mut self.index_handle)
    }


    pub fn salt_store(&mut self, salt:[u8; 16]) -> Result<u32,()>
    // Add a salt to store.
    //
    {
        let size = self.salt_count()? as u32;

        // Update size record
        self.table_salt.seek(SeekFrom::Start(0)).map_err(|_| ())?;
        self.table_salt.write(&pack_u32(size + 1)).map_err(|_| ())?;

        // Write salt to store
        self.table_salt.seek(SeekFrom::End(0)).map_err(|_| ())?;
        self.table_salt.write(&salt).map_err(|_| ())?;

        Ok(size)
    }

    pub fn salt_fetch(&mut self, id:u32) -> Result<[u8; 16],()>
    // Retrieve a salt from store.
    //
    {
        let size = self.salt_count()? as u32;
        let mut buffer = [0u8; 16];
       
        if id < size {
            let offset = 4 + (16 * id as u64);

            self.table_salt.seek(SeekFrom::Start(offset)).map_err(|_| ())?;
            if self.table_salt.read_exact(&mut buffer).is_ok() {
                Ok(buffer)
            } else { Err(()) }
        } else { Err(()) }
    }

    pub fn salt_count(&mut self) -> Result<usize, ()>
    // Get number of salts in store.
    //
    {
        Self::get_header_size(&mut self.table_salt)
    }

    fn get_header_size(file:&mut File) -> Result<usize, ()>
    {
        let mut size_buffer = [0u8; 4];

        file.seek(SeekFrom::Start(0)).map_err(|_| ())?;
        if file.read_exact(&mut size_buffer).is_ok() {
            Ok(unpack_u32(&size_buffer, &mut 0) as usize)
        } else { Err(()) }
    }
}
